import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from chronometry.progress import ProgressBar, iterate
import warnings
from copy import deepcopy
from .InfluenceSimulator import InfluenceSimulator


def q1(x):
	return np.nanquantile(x, 0.25)


def q3(x):
	return np.nanquantile(x, 0.75)


def get_single_column_influence(function, data, x_column, num_deltas):
	if num_deltas % 2 == 1:
		num_deltas += 1

	new_data = deepcopy(data)

	x_mean = new_data[x_column].mean()
	x_sd = new_data[x_column].std()
	x_lower = -x_sd
	x_upper = x_sd
	if x_upper == x_lower:
		if x_mean == 0:
			x_upper = 1
			x_lower = -1
		else:
			x_upper = x_mean * 0.1
			x_lower = -x_mean * 0.1
	step_size = (x_upper - x_lower) / num_deltas
	delta_data = pd.DataFrame({
		'x_delta': list(np.arange(x_lower, 0, step_size)) + list(np.arange(0, x_upper + step_size, step_size))
	})

	new_data['variable'] = x_column
	delta_data['variable'] = x_column
	delta_data['_row_'] = range(len(delta_data))
	new_data = new_data.merge(right=delta_data, on='variable', how='inner')

	test_data = new_data.drop(columns=['x_delta', 'variable', '_row_'])

	new_data['_original_y_'] = function(test_data)
	test_data[x_column] = test_data[x_column] + new_data['x_delta']
	new_data['_new_y_'] = function(test_data)
	new_data['y_delta'] = new_data['_new_y_'] - new_data['_original_y_']
	new_data['delta_ratio'] = new_data['y_delta'] / new_data['x_delta']
	new_data['_count_'] = 1

	aggregate = new_data[['variable', '_row_', '_new_y_', '_original_y_', '_count_', 'y_delta', 'delta_ratio']].groupby(
		['variable', '_row_']
	).agg(
		mean=('_new_y_', np.mean), median=('_new_y_', np.median),
		q1=('_new_y_', q1), q3=('_new_y_', q3),
		min=('_new_y_', np.min), max=('_new_y_', np.max),
		original_mean=('_original_y_', np.mean),
		count=('_count_', np.sum),
		mean_y_delta=('y_delta', np.mean),
		median_y_delta=('y_delta', np.median),
		mean_delta_ratio=('delta_ratio', np.mean),
		median_delta_ratio=('delta_ratio', np.median)
	).reset_index().merge(right=delta_data[['_row_', 'x_delta']], on='_row_', how='left').drop(columns=['_row_'])

	return aggregate


def get_function_influence(function, data, x_columns, num_deltas=200, num_threads=1, echo=1):
	"""
	:param callable function: function that gives a y for a given data
	:param pd.DataFrame data:
	:param list[str] or str x_columns:
	:param int num_deltas:
	:param int num_threads:
	:type echo: int
	:rtype: pd.DataFrame
	"""
	if isinstance(x_columns, str):
		x_columns = [x_columns]
	elif not isinstance(x_columns, list):
		raise TypeError(f'x_columns should be either a string or a list not a {type(x_columns)}')

	def get_influence_data(the_tuple):
		with warnings.catch_warnings():
			warnings.simplefilter('ignore')
			x_column, this_data, this_num_deltas, this_function = the_tuple
			record = get_single_column_influence(
				function=this_function, data=this_data, x_column=x_column, num_deltas=this_num_deltas
			)
		return record

	tuples = [
		(x_column, data, num_deltas, function)
		for x_column in iterate(x_columns, echo=echo, text='preparing data ...')
	]

	progress_bar = ProgressBar(total=len(x_columns) + 1, echo=echo)
	progress_bar.show(amount=len(x_columns), text='data prepared.')

	if num_threads == 1:
		dataframes = [
			get_influence_data(t)
			for t in iterate(iterable=tuples, progress_bar=progress_bar, text='measuring influence ...')
		]
	else:
		dataframes = Parallel(n_jobs=num_threads, backend='threading', require='sharedmem')(
			delayed(get_influence_data)(t)
			for t in iterate(iterable=tuples, progress_bar=progress_bar, text='measuring influence ...')
		)

	result = pd.concat(dataframes).sort_values(by=['variable', 'x_delta'])
	progress_bar.show(amount=len(dataframes) + 1, text='influence measured.')
	return result


def get_model_influence(model, data, x_columns, num_points=200, num_threads=1, echo=1):
	try:
		function = model.predict
	except AttributeError:
		function = model.pred
	'''
	return get_function_influence(
		function=function, data=data, x_columns=x_columns, num_deltas=num_deltas, num_threads=num_threads,
		echo=echo
	)
	'''

	simulator = InfluenceSimulator(
		data=data, function=function, num_threads=num_threads, num_points=num_points
	)
	return simulator.simulate(echo=echo)



from .get_z_score import get_z_score
def get_sample_size(population_size, confidence=0.95, error_margin=0.05, group_proportion=0.5):
	# group_proportion = group_n/sample_n # 0.5 is the worst case scenario
	p = group_proportion
	sample_size = p*(1-p)/((error_margin/get_z_score(confidence=confidence))**2)
	true_sample = (sample_size*population_size)/(sample_size + population_size - 1)
	return true_sample







import re


_first_cap_re = re.compile('(.)([A-Z][a-z]+)')
_all_cap_re = re.compile('([a-z0-9])([A-Z])')
def convert_camel_to_snake(string):
	s1 = _first_cap_re.sub(r'\1_\2', string)
	result = _all_cap_re.sub(r'\1_\2', s1).lower()
	result = re.sub(pattern='_+', repl="_", string=result)
	#print(manipulation, result)
	return result


from .Logit_and_OLS import OLS, Logit
from .GEE import GEE, LogisticGEE


def get_grouped_regression(data, formula, groups, regression_type):
	data = data.copy()
	if isinstance(groups, str):
		data['_groups_'] = data[groups]
	else:
		data['_groups_'] = data.apply(
			axis=1,
			func=lambda x: hash(tuple([
				x[column] for column in groups
			]))
		)
	groups = '_groups_'

	if regression_type == 'linear':
		model = GEE(formula=formula, data=data, groups=groups)
	else:
		model = LogisticGEE(formula=formula, data=data, groups=groups)
	return model

def get_regression_model(formula, data, regression_type, groups, eliminate=False, echo=1):

	if groups is None or groups == []:
		if regression_type == 'linear':
			model = OLS(data=data, formula=formula)
		else:
			model = Logit(data=data, formula=formula)
	else:
		model = get_grouped_regression(
			data=data, formula=formula, groups=groups, regression_type=regression_type
		)

	if not eliminate:
		return model
	else:
		return model.eliminate(num_rounds=None, echo=echo)



from pandas import DataFrame

def convert_simple_table_to_dataframe(simple_table, header=True):
	list_of_lists = simple_table.data

	if header:
		header = list_of_lists[0]
		column_names = header[1:]
		body = list_of_lists[1:]
	else:
		column_names = None
		body = list_of_lists


	index = [l[0] for l in body]
	data = [l[1:] for l in body]
	df = DataFrame.from_records(data=data, columns=column_names, index=index)

	return df



import numpy as np
from sklearn.linear_model import LinearRegression

from .Curve import Curve
EPSILON = 1e-10


class Sigmoid(Curve):
	def __init__(self, x=None, y=None, a=None, b=None, c=None, check_slope=True):
		super().__init__(function=self._sigmoid)
		self._a = a
		self._b = b
		self._c = c
		self._slope = None
		if x is not None and y is not None:
			self.fit(x=x, y=y, check_slope=check_slope)

	def _sigmoid(self, x, a, b, c):
		a = self._a or a
		b = self._b or b
		c = self._c or c


		return a/(np.exp(-b*x + c) + 1)

	def fit(self, x, y, parameter_lower_bounds=None, parameter_upper_bounds=None, check_slope=True):
		x = np.array(x)
		y = np.array(y)

		linear = LinearRegression()
		linear.fit(X=x.reshape((-1, 1)), y=y)
		self._slope = linear.coef_[0]
		if check_slope and self._slope <=0:
			self._fit_successful = False
			return False

		parameter_lower_bounds = parameter_lower_bounds or [0, 0, -np.inf]
		parameter_upper_bounds = parameter_upper_bounds or [np.inf, np.inf, np.inf]
		if self._a:
			parameter_upper_bounds[0] = self._a + EPSILON
			parameter_lower_bounds[0] = self._a - EPSILON
		if self._b:
			parameter_upper_bounds[1] = self._b + EPSILON
			parameter_lower_bounds[1] = self._b - EPSILON
		if self._c:
			parameter_upper_bounds[2] = self._b + EPSILON
			parameter_lower_bounds[2] = self._b - EPSILON

		result = super().fit(x=x, y=y, parameter_lower_bounds=parameter_lower_bounds, parameter_upper_bounds=parameter_upper_bounds)

		if self._a:
			self._optimal_parameters[0] = self._a
		if self._b:
			self._optimal_parameters[1] = self._b
		if self._c:
			self._optimal_parameters[2] = self._c

		return result

	@property
	def a(self):
		return self._optimal_parameters[0]

	@property
	def b(self):
		return self._optimal_parameters[1]

	@property
	def c(self):
		return self._optimal_parameters[2]

	@property
	def parameters(self):
		return {'a': self.a, 'b': self.b, 'c': self.c}



from pandas import DataFrame


def multiply_dictionary(dictionary, key, values):
	"""
	:type dictionary: dict
	:type key: str
	:type values: list
	:rtype: list[dict]
	"""
	result = []
	if not isinstance(values, (list, tuple)):
		values = [values]

	for value in values:
		new_d = dictionary.copy()
		new_d[key] = value
		result.append(new_d)
	return result



def create_grid(dictionary):
	"""
	:param dict[str, list] dictionary: a dictionary of lists of parameter values
	:rtype: list[dict[str,]]
	"""
	result = [{}]
	for key, values in dictionary.items():
		new_result = []
		for dictionary in result:
			new_result += multiply_dictionary(dictionary=dictionary, key=key, values=values)
		result = new_result
	return result


def create_model_grid(model, dictionary, name=None):
	list_of_dictionaries = create_grid(dictionary=dictionary)
	models = {}
	parameters_list = []
	parameters_dictionary = {}
	for index, kwargs in enumerate(list_of_dictionaries):
		model_name = f'{name}_{index}'
		models[model_name] =  model(**kwargs)
		parameters = kwargs.copy()
		parameters['model'] = model_name
		parameters_list.append(parameters)
		parameters_dictionary[model_name] = kwargs.copy()
	return {'models': models, 'parameter_table': DataFrame.from_records(parameters_list), 'parameters': parameters_dictionary}


class ModelGrid:
	def __init__(self, models, grid_dictionaries):
		if not isinstance(models, list):
			models = [models]

		if not isinstance(grid_dictionaries, list):
			grid_dictionaries = [grid_dictionaries]

		num_models_per_type = {}
		out_models = {}
		records = []
		parameter_dictionary = {}
		for model, dictionary in zip(models, grid_dictionaries):
			list_of_dictionaries = create_grid(dictionary=dictionary)

			for index, kwargs in enumerate(list_of_dictionaries):
				try:
					model_type = model.__name__
				except AttributeError:
					model_type = 'model'

				if model_type not in num_models_per_type:
					num_models_per_type[model_type] = 1
				else:
					num_models_per_type[model_type] += 1

				model_name = f'{model_type}_{num_models_per_type[model_type]}'
				out_models[model_name] = model(**kwargs)
				parameters = kwargs.copy()
				parameters['model_name'] = model_name
				records.append(parameters)
				parameter_dictionary[model_name] = kwargs.copy()

		self.models = out_models
		self.num_models_per_type = num_models_per_type
		self.parameters = parameter_dictionary
		self.parameter_table = DataFrame.from_records(records)




from .get_z_score import get_z_score
from .PopulationProportion import get_population_proportion_error, PopulationProportion
from .get_sample_size import get_sample_size
from .MultipleComparisons import MultipleComparisons
from .Nonlinear import Nonlinear, fit_nonlinear
from .Trend import Trend
from .LogarithmicGrowth import LogarithmicGrowth
from .Curve import Curve
from .Sigmoid import Sigmoid
from .perturb import perturb
from .Perturber import Perturber


def get_coefficients(model, columns, raise_error, **kwargs):
	try:
		coefficients = list(model.coef_.flatten())
	except AttributeError:
		coefficients = None

	if coefficients is None:
		return None

	else:
		if len(columns) != len(coefficients):
			if raise_error:
				raise RuntimeError(f'number of columns: {len(columns)}, number of coefficients: {len(coefficients)}')
			else:
				return None
		else:
			coefficient_dictionary = {column: coefficient for coefficient, column in zip(coefficients, columns)}

			for key, value in kwargs.items():
				coefficient_dictionary[key] = value

			return coefficient_dictionary



import pandas as pd
import numpy as np
from ravenclaw import get_numeric_columns
from joblib import Parallel, delayed
from chronometry.progress import iterate, ProgressBar
from copy import deepcopy


def get_chunk_indices(num_indices, chunk_size):
	chunk_size = max(1, chunk_size)
	return [(i, min(i+chunk_size, num_indices)) for i in range(0, num_indices+1, chunk_size)]


def q1(x):
	return np.nanquantile(x, 0.25)


def q3(x):
	return np.nanquantile(x, 0.75)


class InfluenceSimulator:
	def __init__(self, data, function, num_threads=1, num_points=200):
		"""
		:type data: pd.DataFrame
		:type function: callable
		:type num_threads: int
		:type num_points: int
		"""
		self._data = data
		self._numeric_column_names = get_numeric_columns(data=data)
		self._num_threads = num_threads
		self._means = None
		self._sds = None
		self._function = function
		self._num_points = num_points + 1 if num_points % 2 == 1 else num_points
		self._simulation_points = None
		self._influencers = {}
		# self._memory = {}
		self._original_output = self._function(data)
		self._calculate_statistics()

	def get_influencers(self, column_name):
		upper = self._sds[column_name]
		if upper == 0:
			if self._means[column_name] == 0:
				lower, upper = -1, 1
			else:
				lower, upper = -0.1 * self._means[column_name], 0.1 * self._means[column_name]
		else:
			lower = -upper
		step_size = (upper - lower) / self._num_points
		return list(np.arange(lower, 0, step_size)) + list(np.arange(0, upper + step_size, step_size))

	@property
	def parallel_process(self):
		return Parallel(n_jobs=self._num_threads, backend='threading', require='sharedmem')

	@property
	def numeric_columns(self):
		"""
		:rtype: list[pd.Series]
		"""
		return [self._data[column] for column in self._numeric_column_names]

	def _calculate_statistics(self):
		output = self._function(self._data)
		self._output_mean = np.nanmean(output)
		self._output_sd = np.nanstd(output)

		if self._num_threads == 1:

			self._means = {np.nanmean(column) for column in self.numeric_columns}
			self._sds = {np.nanstd(column) for column in self.numeric_columns}
			self._influencers = {
				column_name: self.get_influencers(column_name=column_name)
				for column_name in self._numeric_column_names
			}

		else:
			means = self.parallel_process(
				delayed(np.nanmean)(column) for column in self.numeric_columns
			)
			self._means = {name: mean for name, mean in zip(self._numeric_column_names, means)}

			sds = self.parallel_process(
				delayed(np.nanstd)(column) for column in self.numeric_columns
			)
			self._sds = {name: sd for name, sd in zip(self._numeric_column_names, sds)}

			influence_lists = self.parallel_process(
				delayed(self.get_influencers)(column) for column in self._numeric_column_names
			)
			self._influencers = {
				column_name: influence_list
				for column_name, influence_list in zip(self._numeric_column_names, influence_lists)
			}

	def get_single_influence(self, column, perturbation, function):
		new_data = self._data.drop(columns=column)
		new_data[column] = self._data[column] + perturbation
		new_data = new_data[self._data.columns]
		output = function(new_data)
		difference = output - self._original_output
		return {
			'y_mean': np.nanmean(output), 'y_sd': np.nanstd(output),
			'y_q1': q1(output), 'y_median': np.median(output), 'y_q3': q3(output),
			'influence_mean': np.nanmean(difference), 'influence_sd': np.nanstd(difference),
			'influence_q1': q1(difference), 'influence_median': np.median(difference), 'influence_q3': q3(difference)
		}

	def simulate(self, echo=1):
		"""
		:type echo: int
		:rtype: list[dict]
		"""
		def _get_influence(column, perturbation, function):
			influence_result = self.get_single_influence(column=column, perturbation=perturbation, function=function)
			influence_result['column'] = column
			influence_result['perturbation'] = perturbation
			return influence_result

		progress_bar = ProgressBar(total=2 + len(self._influencers), echo=echo)

		if self._num_threads == 1:
			result = [
				_get_influence(column=column, perturbation=perturbation, function=self._function)
				for column, influences in iterate(self._influencers.items(), progress_bar=progress_bar)
				for perturbation in influences
			]
		else:
			result = self.parallel_process(
				delayed(_get_influence)(column=column, perturbation=perturbation, function=deepcopy(self._function))
				for column, influences in iterate(self._influencers.items(), progress_bar=progress_bar)
				for perturbation in influences
			)

		progress_bar._total = 2 + len(self._influencers)
		progress_bar.show(amount=1 + len(self._influencers))

		result_data = pd.DataFrame.from_records(result)
		progress_bar.show(amount=2 + len(self._influencers))
		return result_data



from sklearn.linear_model import LinearRegression
from pandas import DataFrame
from pensieve import Pensieve


class Layer:
	def __init__(self, models, name_numbers=None, pensieve=None, interaction=True):
		"""
		:type models: dict[str, LinearRegression] or list[LinearRegression]
		:type name_numbers: dict[str, int] or NoneType
		:type pensieve: Pensieve or NoneType
		"""
		if name_numbers is None:
			name_numbers = {}
		if isinstance(models, (list, tuple)):
			models_dictionary = {}
			for model in models:
				name = type(model).__name__
				if name not in name_numbers:
					name_numbers[name] = 1
				else:
					name_numbers[name] += 1
				models_dictionary[f'{name}_{name_numbers[name]}'] = model
			models = models_dictionary

		self._models = models
		self._interaction = interaction
		self._pensieve = pensieve

	@property
	def models(self):
		"""
		:rtype: dict[str, LinearRegression]
		"""
		return self._models

	@property
	def pensieve(self):
		"""
		:rtype: Pensieve or NoneType
		"""
		return self._pensieve

	def fit(self, X, y):
		for model_name, model in self.models.items():
			model.fit(X=X, y=y)
			if self.pensieve is not None:
				self.pensieve[model_name] = model

	@staticmethod
	def _predict_with_models(models_dictionary, X, interaction, last_layer=False):

		def predict_or_transform(model, X):
			try:
				return model.predict(X=X)
			except AttributeError:
				return DataFrame(model.transform(X=X))

		if len(models_dictionary) > 1 or not last_layer:
			data_dictionary = {}
			for model_name, model in models_dictionary.items():
				prediction = predict_or_transform(model=model, X=X)
				if isinstance(prediction, DataFrame):
					for column_num, column in enumerate(prediction.columns):
						data_dictionary[f'{model_name}_prediction_{column_num + 1}'] = prediction[column]
				else:
					data_dictionary[f'{model_name}_prediction'] = prediction

			if interaction:
				new_dictionary = {}
				for key1, value1 in data_dictionary.items():
					for key2, value2 in data_dictionary.items():
						new_dictionary[f'{key1}__{key2}'] = value1 * value2
			data_dictionary.update(new_dictionary)
			return DataFrame(data_dictionary)

		else:
			return models_dictionary[list(models_dictionary.keys())[0]].predict(X=X)

	def predict(self, X, x_name='X', y_name='y', materialize=True, last_layer=False):
		if self.pensieve is not None:
			if X is not None:
				self.pensieve[x_name] = X

			def create_predict_function(models, x_name, last_layer):
				def predict_function(x):
					return self._predict_with_models(
						models_dictionary={model_name: x[model_name] for model_name, model in models.items()},
						X=x[x_name],
						last_layer=last_layer, interaction=self._interaction
					)
				return predict_function
			self.pensieve.store(
				key=y_name, precursors=[x_name] + list(self.models.keys()),
				function=create_predict_function(models=self.models, x_name=x_name, last_layer=last_layer),
				materialize=materialize
			)
			return self.pensieve[y_name]

		else:
			return self._predict_with_models(
				models_dictionary=self.models, X=X, last_layer=last_layer, interaction=self._interaction
			)


class Ensemble:
	def __init__(self, layers, pensieve=None, interaction=True):
		"""
		:type layers: list[dict[str, LinearRegression]] or list[list[LinearRegression]]
		:type pensieve: Pensieve or NoneType
		"""
		name_numbers = {}
		self._layers = []
		self._pensieve = pensieve
		self._interaction = interaction
		for models in layers:
			self._layers.append(
				Layer(models=models, name_numbers=name_numbers, pensieve=self.pensieve, interaction=self._interaction)
			)

	@property
	def layers(self):
		"""
		:type: list[Layer]
		"""
		return self._layers

	@property
	def output_model(self):
		"""
		:rtype: LinearRegression
		"""
		return self._output_model

	@property
	def pensieve(self):
		"""
		:rtype: Pensieve or NoneType
		"""
		return self._pensieve

	def fit(self, X, y):
		for layer in self.layers:
			layer.fit(X=X, y=y)
			X = layer.predict(X=X, last_layer=False)
			#print(X.shape)

	def predict(self, X):
		for layer_number, layer in enumerate(self.layers):
			if layer_number == 0:
				x_name = 'X'
			else:
				x_name = f'X_{layer_number + 1}'

			if layer_number + 1 == len(self.layers):
				y_name = 'y'
			else:
				y_name = f'X_{layer_number + 2}'

			X = layer.predict(
				X=X if layer_number==0 or self.pensieve is None else None,
				x_name=x_name, y_name=y_name, last_layer=layer_number + 1 == len(self.layers)
			)
			#print(X.shape)

		return X



from sklearn.metrics import confusion_matrix
import numpy as np
from math import sqrt


def evaluate_classification(actual, predicted):
	actual = np.array(actual)
	predicted = np.array(predicted)

	cm = confusion_matrix(actual, predicted)

	try:
		true_negative, false_positive, false_negative, true_positive = cm.ravel()
	except:
		print(actual, type(actual), predicted, type(predicted), cm, cm.ravel())
		raise

	positive = false_negative + true_positive
	negative = true_negative + false_positive
	sensitivity = true_positive/positive
	specificity = true_negative/negative
	precision = true_positive/(true_positive + false_positive)
	negative_predictive_value = true_negative/(true_negative + false_negative)
	miss_rate = false_negative/positive
	fall_out = false_positive/negative
	false_discovery_rate = false_positive/(false_positive + true_positive)
	false_omission_rate = false_negative/(false_negative + true_negative)
	threat_score = true_positive/(true_positive + false_negative + false_positive)
	accuracy = (true_positive + true_negative)/(positive + negative)
	f1_score = 2*true_positive/(2*true_positive + false_positive + false_negative)
	matthews_correlation_coefficient = (true_positive * true_negative - false_positive * false_negative) / sqrt(
		(true_positive + false_positive) * (true_positive + false_negative) * (true_negative + false_positive) * (true_negative + false_negative)
	)
	informedness = sensitivity + specificity - 1
	markedness = precision + negative_predictive_value - 1

	return {
		'accuracy': accuracy,
		'precision': precision,
		'recall': sensitivity,
		'specificity': specificity,

		'negative_predictive_value': negative_predictive_value,
		'miss_rate': miss_rate,
		'fall_out': fall_out,
		'false_discovery_rate': false_discovery_rate,
		'false_omission_rate': false_omission_rate,
		'threat_score': threat_score,
		'f1_score': f1_score,
		'matthews_correlation_coefficient': matthews_correlation_coefficient,
		'informedness': informedness,
		'markedness': markedness,
		'confusion_matrix': cm,
		'true_negative': true_negative,
		'false_positive': false_positive,
		'false_negative': false_negative,
		'true_positive': true_positive
	}



from pandas import DataFrame, concat
from numpy import nan, isnan
from statsmodels.api import OLS, GEE, Logit
from statsmodels.api import families
from ravenclaw.wrangling import bring_to_front
from chronometry.progress import ProgressBar
import re

from .effect import Formula, MainEffect, InteractionEffect, Effect
from .convert_simple_table_to_dataframe import convert_simple_table_to_dataframe
from .convert_camel_to_snake import convert_camel_to_snake
from .exceptions import BrokenModel, BrokenSummary, FormulaError


class Regression:
	def __init__(
			self, data, model_builder, formula, significance_level=0.05, family=None, groups=None, parent=None
	):
		"""
		:type data: DataFrame
		:type model_builder: callable
		:type formula: str or Formula
		:type significance_level: float
		:type family: families.Binomial or families.Family or NoneType
		:type groups: NoneType or str or list[str]
		:type previous_formulas: NoneType or list[Formula]
		"""
		self._formula = Formula(formula)
		if self._formula.dependent_variable is None:
			raise FormulaError(f'dependent variable missing from {self._formula}!')
		self._groups = groups
		self._data = data
		self._significance_level = significance_level
		self._model_builder = model_builder
		self._family = family
		self._groups = groups

		self._variables = {effect.name: effect for effect in self._formula.effects}
		self._model = None
		self._fit = None
		self._summary = None
		self._summary_table = None
		self._parent = parent
		self._effects = None
		self._significant_interaction_effects = None
		self._significant_main_effects = None
		self._insignificant_interaction_effects = None
		self._insignificant_main_effects = None

	def __repr__(self):
		return f'formula: {self.formula.represent()}\ngroups: {self.groups}\nfamily: {self.family}'

	def __str__(self):
		return self.__repr__()

	@property
	def formulas(self):
		"""
		:rtype: list[Formula]
		"""
		if self._parent is None:
			return [self.formula]
		else:
			return self._parent.formulas + [self.formula]

	def display(self, p=None):
		try:
			from IPython.core.display import display
			display((
				self._model_builder,
				{
					'formula': self.formula,
					'groups': self.groups,
					'family': self.family,
					'summary_table': self._summary_table
				},
				self.data.head()
			))
		except ImportError:
			if p is not None:
				p.pretty(self.__repr__())
			else:
				print(self.__repr__())

	def _repr_pretty_(self, p, cycle):
		if cycle:
			p.text('Regression')
		else:
			self.display(p=p)

	def build_model(self):
		"""
		:rtype: OLS or GEE or Logit
		"""
		if self.family is None and self.groups is None:
			model = self._model_builder(formula=self.formula.formula, data=self.data)
		elif self.family is None:
			model = self._model_builder(formula=self.formula.formula, data=self.data, groups=self.groups)
		elif self.groups is None:
			model = self._model_builder(formula=self.formula.formula, data=self.data, family=self.family)
		else:
			model = self._model_builder(
				formula=self.formula.formula, data=self.data, groups=self.groups, family=self.family
			)
		return model

	@property
	def variables(self):
		"""
		:rtype: dict[str, MainEffect or InteractionEffect]
		"""
		return self._variables

	@property
	def model(self):
		"""
		:rtype: OLS or GEE or Logit
		"""
		if self._model is None:
			self._model = self.build_model()
		return self._model

	@property
	def fit(self):
		"""
		:rtype: OLS or GEE or Logit
		"""
		if self._fit is None:
			try:
				self._fit = self.model.fit()
			except Exception as e:
				self._fit = BrokenModel(exception=e, regression=self)

		return self._fit

	def predict(self, data):
		return self.fit.predict(data)

	@property
	def summary(self):
		if self._summary is None:
			self._summary = self.fit.summary()
		return self._summary

	@property
	def coefficient_table(self):
		"""
		:rtype: DataFrame
		"""
		if isinstance(self.summary, BrokenSummary):
			return DataFrame({'error': [self.summary.model.exception]})
		else:
			return convert_simple_table_to_dataframe(self.summary.tables[1]).rename(columns={
				'coef': 'coefficient', 'std err': 'standard_error', '[0.025': 'lower_0.025',
				'0.975]': 'upper_0.975', 'P>|z|': 'p'
			})

	@property
	def model_table(self):
		"""
		:rtype: DataFrame
		"""
		result = convert_simple_table_to_dataframe(self.summary.tables[2], header=False).reset_index()
		tables = []
		for i in range(result.shape[1]//2):
			table = result.iloc[:, i*2:i*2+2]
			table.columns = ['name', 'value']
			tables.append(table)
		result = concat(tables).reset_index(drop=True)

		def clean_variable_name(x):
			x = x.strip()
			x = re.sub(':\s*$', '', x)
			x = re.sub('-', ' ', x)
			x = re.sub('[^a-zA-Z\d\s:]', ' ', x)
			x = '_'.join(x.split())

			x = convert_camel_to_snake(x)
			x = re.sub('\s+', '', x)
			return x

		result['name'] = result['name'].apply(clean_variable_name)
		return result.set_index(keys='name')

	@property
	def summary_table(self):
		if self._summary_table is None:
			table = self.coefficient_table

			if not isinstance(self.summary, BrokenSummary):
				for column in table.columns:
					if column != 'coefficient':
						table[column] = table[column].astype(float)
				table['variable_formula'] = table.index
				table['p'] = table['p']

				effects = table.apply(
					func=lambda x: Effect(formula=x['variable_formula'], p=x['p'], coefficient=x['coefficient']),
					axis=1
				)
				table['num_interactions'] = effects.apply(func=lambda x: x.num_interactions)

				max_num_interactions = table['num_interactions'].max()

				variables = []

				def get_name(effect):
					if effect is None:
						return None
					else:
						return effect.name

				for i in range(max_num_interactions + 1):
					variable = f'variable_{i+1}'
					table[variable] = effects.apply(func=lambda x: get_name(x.get_variable(i, missing=None)))
					variables.append(variable)

				table['effect'] = effects
				table['effect_type'] = effects.apply(func=lambda x: x.type)
				table['categorical_level'] = effects.apply(func=lambda x: x.categorical_level)
				table['relative'] = effects.apply(func=lambda x: x.relative)

				table['effect_name'] = effects.apply(func=lambda x: x.name)

				table['minimum_variable_p'] = table.groupby(by='effect')['p'].transform('min')

				table['significant'] = table['minimum_variable_p'] < self.significance_level

				# get other levels for all categorical
				base_levels = table[
					(table['effect_type'] == 'categorical') & (table['relative'])
					][['effect', 'categorical_level']].groupby(
					by='effect'
				).apply(lambda x: list(set(self.data[x['effect'][0]]) - set(x['categorical_level']))[0]).reset_index().rename(
					columns={0: 'base_level'}
				)
				table = table.merge(right=base_levels, on='effect', how='left')

				table['is_intercept'] = table['effect'] == 'intercept'
				table.sort_values(by=['is_intercept', 'minimum_variable_p'], ascending=[False, True], inplace=True)
				if 'base_level' not in table.columns:
					table['base_level'] = None
				table = table.replace({nan: None})
				self._summary_table = bring_to_front(
					data=table.drop(
						columns=['minimum_variable_p', 'is_intercept', 'relative']
					).reset_index(drop=True),
					columns=['effect_type', 'effect', 'num_interactions'] + variables + ['categorical_level', 'base_level', 'significant', 'p']
				)
			else:
				self._summary_table = DataFrame({
					column: [None] for column in [
						'effect_type', 'effect', 'num_interactions', 'categorical_level', 'base_level', 'significant', 'p'
					]
				})
				self._summary_table['error'] = self.summary.model.exception

		return self._summary_table

	@property
	def family(self):
		"""
		:rtype: families.Binomial or families.Family or NoneType
		"""
		return self._family

	@property
	def groups(self):
		"""
		:rtype: NoneType or str or list[str]
		"""
		return self._groups

	@property
	def formula(self):
		"""
		:rtype: Formula
		"""
		return self._formula

	@property
	def data(self):
		"""
		:rtype: DataFrame
		"""
		return self._data

	@property
	def significance_level(self):
		"""
		:rtype: float
		"""
		return self._significance_level

	@property
	def except_intercept(self):
		return self.summary_table[self.summary_table['effect_type'] != 'intercept']

	@property
	def num_significant_effects(self):
		try:
			return len(self.except_intercept[self.except_intercept['significant']])
		except TypeError as e:
			if len(self.except_intercept) < 2:
				return None
			else:
				print(self.except_intercept)
				raise e

	@property
	def num_insignificant_effects(self):
		return len(self.insignificant_effects)

	@property
	def effects(self):
		if self._effects is None:
			self._effects = sorted(self.except_intercept['effect'], key=lambda x: x.p)
		return self._effects

	@property
	def significant_interaction_effects(self):
		"""
		:rtype: list[InteractionEffect]
		"""
		if self._significant_interaction_effects is None:
			self._significant_interaction_effects = [
				effect for effect in self.effects
				if isinstance(effect, InteractionEffect) and (effect.p < self._significance_level or isnan(effect.p) or effect.p is None)
			]
		return self._significant_interaction_effects

	@property
	def insignificant_interaction_effects(self):
		"""
		:rtype: list[InteractionEffect]
		"""
		if self._insignificant_interaction_effects is None:
			self._insignificant_interaction_effects = [
				effect for effect in self.effects
				if isinstance(effect, InteractionEffect) and effect.p >= self._significance_level and not isnan(effect.p) and effect.p is not None
			]
		return self._insignificant_interaction_effects

	def is_main_effect_significant(self, main_effect):
		"""
		:type main_effect: MainEffect
		:rtype: bool
		"""
		return any([
			interaction.contains(main_effect) for interaction in self.significant_interaction_effects
		]) or main_effect.p < self.significance_level

	@property
	def significant_main_effects(self):
		"""
		:rtype: list[MainEffect]
		"""
		if self._significant_main_effects is None:
			self._significant_main_effects = [
				effect for effect in self.effects
				if isinstance(effect, MainEffect) and self.is_main_effect_significant(effect)
			]
		return self._significant_main_effects

	@property
	def insignificant_main_effects(self):
		"""
		:rtype: list[MainEffect]
		"""
		if self._insignificant_main_effects is None:
			self._insignificant_main_effects = [
				effect for effect in self.effects
				if isinstance(effect, MainEffect) and not self.is_main_effect_significant(effect)
			]
		return self._insignificant_main_effects

	@property
	def insignificant_effects(self):
		"""
		:rtype: list[MainEffect or InteractionEffect]
		"""
		return self.insignificant_main_effects + self.insignificant_interaction_effects

	@property
	def least_significant_insignificant_effect(self):
		return self.insignificant_effects[-1]

	@property
	def all_effects_but_least_significant(self):
		return [effect for effect in self.effects if effect != self.least_significant_insignificant_effect]

	def eliminate(self, num_rounds=None, echo=1):
		"""
		:type num_rounds: int or NoneType
		:type echo: bool or int
		:rtype: Regression
		"""
		model = self
		if isinstance(model.fit, BrokenModel):
			return self

		num_rounds = num_rounds or len(self.variables)
		progress_bar = ProgressBar(total=num_rounds, echo=echo)
		round = 0
		previous_lse = None
		previous_formula = None
		for round in range(num_rounds):

			if model.num_insignificant_effects < 1:
				break

			else:
				y = model.formula.dependent_variable


				lse = model.least_significant_insignificant_effect
				xs = model.all_effects_but_least_significant
				progress_bar.show(
					amount=round,
					text=f'effects: {len(xs)+1}, eliminating {lse} with p={lse.p}'
				)
				new_formula = Formula.from_variables(
					independent_variables=xs,
					dependent_variable=y
				)

				if previous_lse == lse:
					print('eliminating', lse, '\n\n')
					print('is main effect insignificant?', model.is_main_effect_significant(lse), '\n\n')
					for interaction in model.significant_interaction_effects:
						print(interaction, 'contains?', interaction.contains(lse))

					for interaction in model.insignificant_interaction_effects:
						print(interaction, interaction.p)

					print(model.least_significant_insignificant_effect, lse)
					print('\n\ninsignificant effects\n', model.insignificant_effects)
					print(new_formula.display())

					for x in xs:
						print(x, x.num_interactions, x.p)
					raise RuntimeError(f'repeating the elimination of {lse}')

				previous_formula = new_formula
				previous_lse = lse


				new_model = model.__class__(
					data=model.data, formula=new_formula, significance_level=model.significance_level,
					groups=model.groups, family=model.family, model_builder=model._model_builder,
					parent=model
				)



				if isinstance(new_model.fit, BrokenModel):
					break

				else:
					model = new_model

		progress_bar.show(
			amount=round + 1,
			text=f'effects: {len(model.formula.independent_variables)}'
		)
		return model



import numpy as np
from scipy.optimize import curve_fit
from slytherin.collections import Series


class Curve:
	def __init__(self, function):
		self._optimal_parameters = None
		self._parameter_covariance = None
		self._fit_successful = None
		self._function = function
		self._curve = None

	def fit(self, x, y, parameter_lower_bounds=None, parameter_upper_bounds=None):
		x = np.array(x)
		y = np.array(y)

		parameter_lower_bounds = parameter_lower_bounds or -np.inf
		parameter_upper_bounds = parameter_upper_bounds or np.inf

		try:
			self._optimal_parameters, self._parameter_covariance = curve_fit(
				self._function, x, y, bounds=(parameter_lower_bounds, parameter_upper_bounds)
			)
			self._curve = lambda x: self._function(x, *self.optimal_parameters)
			self._fit_successful = True

		except Exception as e:
			self._fit_successful = False
			raise e

		return self._fit_successful

	@property
	def optimal_parameters(self):
		return self._optimal_parameters

	def predict(self, x):
		if self._fit_successful is None:
			raise RuntimeError('Curve is not fit yet!')

		elif not self._fit_successful:
			raise RuntimeError('Curve fit was unsuccessful!')

		else:
			return_type = 'list' if isinstance(x, list) else 'not_list'
			x = np.array(x)
			y = self._curve(x)
			if return_type == 'list':
				y = list(y)
			return y

	def plot(self, x, y=None):
		try:
			import matplotlib.pyplot as plt
		except ModuleNotFoundError:
			print('matplotlib is not installed')
			return False

		smooth_x = Series(start=min(x), end=max(x), step=(max(x) - min(x))/(10*len(x))).list
		plt.plot(smooth_x, self.predict(smooth_x), color='lightcoral')
		if y is not None:
			plt.plot(x, y, 'ro', color='navy')
		else:
			plt.plot(x, self.predict(x), 'ro', color='lightcoral')



from .MainEffect import MainEffect


class InteractionEffect:
	def __init__(self, formula, p=None, coefficient=None):
		formula = formula.replace(':', "*")
		self._main_effects = sorted([MainEffect(formula=x) for x in formula.split('*')], key=lambda x: x.formula)
		self._p = p
		self._coefficient = coefficient

	def __eq__(self, other):
		if not isinstance(other, self.__class__):
			return False
		elif self.num_interactions != other.num_interactions:
			return False
		return self.formula == other.formula

	def __lt__(self, other):
		return (self.num_interactions, self.name) < (other.num_interactions, other.name)

	def __gt__(self, other):
		return (self.num_interactions, self.name) > (other.num_interactions, other.name)

	def __le__(self, other):
		return (self.num_interactions, self.name) <= (other.num_interactions, other.name)

	def __ge__(self, other):
		return (self.num_interactions, self.name) >= (other.num_interactions, other.name)

	def __hash__(self):
		return hash(self.name)

	def contains(self, main_effect):
		for x in self.main_effects:
			if x.formula == main_effect.formula:
				return True
		return False

	@property
	def main_effects(self):
		"""
		:rtype: list[MainEffect]
		"""
		return self._main_effects

	@property
	def coefficient(self):
		return self._coefficient

	@property
	def p(self):
		return self._p

	@property
	def num_interactions(self):
		return len(self.main_effects) - 1

	def get_variable(self, index, missing=None):
		if index < len(self.main_effects):
			return self.main_effects[index]
		else:
			return missing

	@property
	def type(self):
		"""
		:rtype: str
		"""
		return '-'.join(sorted([main_effect.type for main_effect in self.main_effects]))

	@property
	def formula(self):
		"""
		:rtype: str
		"""
		return self.represent(multiply_symbol='*')

	def represent(self, multiply_symbol='×'):
		"""
		:rtype: str
		"""
		return f' {multiply_symbol} '.join([main_effect.formula for main_effect in self.main_effects])

	@property
	def name(self):
		"""
		:rtype: str
		"""
		return ' × '.join([main_effect.name for main_effect in self.main_effects])

	def __repr__(self):
		return self.name

	def __str__(self):
		return self.__repr__()

	@property
	def categorical_level(self):
		return None

	@property
	def relative(self):
		return None




import re


class MainEffect:
	def __init__(self, formula, p=None, coefficient=None):
		self._formula = formula.strip()
		self._p = p
		self._coefficient = coefficient
		search_result = re.search('^C\((.*)\)\[(.*)\]$', self._formula)
		self._level = None
		self._relative = None
		if search_result:
			self._type = 'categorical'
			self._name = search_result.group(1)
			self._level = search_result.group(2)
			if self._level.startswith('T.'):
				self._level = self._level[2:]
				self._relative = True

			else:
				self._relative = False

		elif formula.lower() == 'intercept':
			self._type = 'intercept'
			self._name = 'intercept'

		else:
			self._type = 'numeric'
			self._name = self._formula

	def __eq__(self, other):
		if not isinstance(other, self.__class__):
			return False
		return self.formula == other.formula

	def __lt__(self, other):
		return self.name < other.name

	def __gt__(self, other):
		return self.name > other.name

	def __le__(self, other):
		return self.name <= other.name

	def __ge__(self, other):
		return self.name >= other.name

	def __hash__(self):
		return hash(self.name)

	def __str__(self):
		return self.name

	def __repr__(self):
		return self.__str__()

	@property
	def coefficient(self):
		return self._coefficient

	@property
	def p(self):
		return self._p

	@property
	def num_interactions(self):
		return 0

	def get_variable(self, index, missing=None):
		if index == 0:
			return self
		else:
			return missing

	@property
	def formula(self):
		return self._formula

	def represent(self, multiply_symbol=''):
		return self.formula

	@property
	def type(self):
		return self._type

	@property
	def name(self):
		return self._name

	@property
	def categorical_level(self):
		return self._level

	@property
	def relative(self):
		return self._relative



import pandas as pd
from random import sample, choice, random
from numpy.random import normal


def create_xy_data(
		num_rows=100, num_x_columns=10, correlation_ratio=0.5, correlation_noise=0.1, num_coefficients=5,
		interactions=True
):
	num_correlating_features = round(correlation_ratio * num_x_columns)
	num_independent_features = num_x_columns - num_correlating_features

	independent_columns = [f'x{i}' for i in range(1, num_independent_features + 1)]

	data = pd.DataFrame({
		column: normal(scale=1, size=num_rows)
		for column in independent_columns
	})

	correlating_columns = [f'x{i}' for i in range(num_independent_features + 1, num_x_columns + 1)]

	for column in correlating_columns:
		independent_column = choice(independent_columns)
		print(f'{column} correlates with {independent_column}')
		data[column] = data[independent_column] + normal(scale=correlation_noise, size=num_rows)

	coefficients = {
		column: random()
		for column in sample(independent_columns + correlating_columns, k=num_coefficients)
	}

	if interactions:
		interaction_coefficients = {
			(column_1, column_2): random() * 10
			for column_1 in sample(independent_columns + correlating_columns, k=round(num_coefficients ** 0.5))
			for column_2 in sample(independent_columns + correlating_columns, k=round(num_coefficients ** 0.5))
		}
	else:
		interaction_coefficients = {}

	data['y'] = normal(scale=correlation_noise * num_coefficients, size=num_rows)

	parts = []

	for column, coefficient in coefficients.items():
		data['y'] = data['y'].values + data[column].values * coefficient
		parts.append(f'{round(coefficient, 2)} {column}')

	for columns, coefficient in interaction_coefficients.items():
		column_1, column_2 = columns
		data['y'] = data['y'].values + data[column_1].values * data[column_2].values * coefficient
		parts.append(f'{round(coefficient, 2)} {column_1}.{column_2}')

	print(f'y = {" + ".join(parts)}')

	return data











import statsmodels.formula.api as smf
from .exceptions import BrokenModel
from .Regression import Regression


class Logit(Regression):

	def __init__(
			self, data, formula, significance_level=0.05, model_builder=None, family=None, groups=None, parent=None
	):
		super().__init__(
			data=data, formula=formula, model_builder=smf.logit, significance_level=significance_level, parent=parent
		)

	@property
	def coefficient_table(self):
		"""
		:rtype: DataFrame
		"""
		return super().coefficient_table.rename(columns={'P>|z|': 'p'})

	@property
	def fit(self):
		"""
		:rtype: Logit
		"""
		if self._fit is None:
			try:
				self._fit = self.model.fit(disp=False)
			except Exception as e:
				self._fit = BrokenModel(exception=e, regression=self)

		return self._fit

	@property
	def name(self):
		"""
		:rtype: str
		"""
		return 'logit'

class OLS(Regression):

	def __init__(self, data, formula, significance_level=0.05, model_builder=None, family=None, groups=None, parent=None):
		super().__init__(data=data, formula=formula, model_builder=smf.ols, significance_level=significance_level, parent=parent)

	@property
	def fit(self):
		"""
		:rtype: OLS
		"""
		if self._fit is None:
			try:
				self._fit = self.model.fit(disp=False)
			except Exception as e:
				self._fit = BrokenModel(exception=e, regression=self)

		return self._fit

	@property
	def name(self):
		"""
		:rtype: str
		"""
		return 'ols'

	@property
	def coefficient_table(self):
		"""
		:rtype: DataFrame
		"""
		return super().coefficient_table.rename(columns={'P>|t|': 'p'})



from ravenclaw import get_numeric_columns, is_numeric
from chronometry.progress import ProgressBar
from pandas import concat
from pensieve import Pensieve


def perturb_numeric_column(data, column, delta):
	data = data.copy()
	data['perturbed_column'] = column
	data['original_value'] = data[column].values
	data['perturbation'] = delta
	data[column] = data['original_value'] + delta
	return data

def perturb_categorical_column(data, column, category):
	data = data.copy()
	data['perturbed_column'] = column
	data['original_value'] = data[column].values
	data['perturbation'] = data[column].apply(lambda x: f'{x} -> {category}')
	data[column] = category
	return data

def perturb(data, numeric_perturbation, columns=None, transformation=None, function=None, echo=1):
	"""
	perturbs a dataframe by adding and subtracting delta from one or all numeric columns
	:type data: pandas.DataFrame
	:type numeric_perturbation: float or list[float] or dict[str, float] or dict[str, list[float]]
	:type columns: NoneType or str or list[str]
	:type echo: bool or int or interaction.ProgressBar
	:type transform: NoneType or callable
	:type function: NoneType or callable
	:return:
	"""
	pensieve = Pensieve()
	try:

		data = data.copy()
		data['perturbation_id'] = range(len(data))
		pensieve['data'] = data

		pensieve['transformation'] = transformation

		pensieve['input_columns'] = columns
		pensieve['numeric_columns'] = lambda data, input_columns: [
			column
			for column in (input_columns or data.columns)
			if is_numeric(data[column])
		]
		pensieve['categorical_columns'] = lambda data, input_columns: [
			column
			for column in (input_columns or data.columns)
			if not is_numeric(data[column])
		]

		pensieve['input_perturbation'] = numeric_perturbation
		def get_perturbation(input_perturbation, columns):
			if not isinstance(input_perturbation, dict):
				perturbation_dictionary = {column: input_perturbation for column in columns}
			else:
				perturbation_dictionary = {column: input_perturbation[column] for column in columns if column in input_perturbation}

			return {
				key: value if isinstance(value, (list, tuple, set)) else [value]
				for key, value in perturbation_dictionary.items()
			}
		pensieve.store(key='perturbation', function=get_perturbation)

		def perturb_data(data, perturbation):
			list_of_data = []
			progress_bar = ProgressBar(total=len(perturbation), echo=echo)
			progress = 0
			for column, perturbation_list in perturbation.items():
				progress_bar.show(amount=progress, text=column)
				for delta in perturbation_list:
					list_of_data.append(perturb_numeric_column(data=data, column=column, delta=delta))
				progress += 1
			data = concat(list_of_data)
			progress_bar.show(amount=progress)
			return data
		pensieve.store(key='perturbed_data', function=perturb_data)

		def transform_data(data, transformation):
			data = data.copy()
			if transformation is None:
				def transformation(x):
					return x
			result = transformation(data.drop(columns='perturbation_id'))
			if len(result) != len(data):
				raise RuntimeError(f'transformation changes data length from {len(data)} to {len(result)}')
			result['perturbation_id'] = data['perturbation_id']
			return result
		pensieve.store(key='data__transformed', function=transform_data)

		pensieve.store(
			key='perturbed_data__transformed',
			function=lambda perturbed_data, transformation: transform_data(
				data=perturbed_data, transformation=transformation
			)
		)

		pensieve.store(
			key='original_columns',
			function=lambda data__transformed: [
				column for column in data__transformed.columns if column != 'perturbation_d'
			]
		)

		pensieve['function'] = function

		def apply_function(data, function, original_columns):
			result = data.copy()
			if function is not None:
				result['function_result'] = function(result[original_columns])
			return result
		pensieve.store(
			key='function_result',
			function=lambda data__transformed, function, original_columns: apply_function(
				data=data__transformed,
				function=function,
				original_columns=original_columns
			)
		)
		pensieve.store(
			key='perturbed__function_result',
			function=lambda perturbed_data__transformed, function, original_columns: apply_function(
				data=perturbed_data__transformed,
				function=function,
				original_columns=original_columns
			)
		)

		def compare(perturbed__function_result, function_result):
			perturbed__function_result = perturbed__function_result.copy()
			if 'function_result' in function_result.columns:
				function_result = function_result[['perturbation_id', 'function_result']]
				function_result = function_result.rename(columns={'function_result': 'original_function_result'})

				perturbed__function_result = perturbed__function_result.rename(
					columns={'function_result': 'perturbed_function_result'}
				)
				comparison = perturbed__function_result.merge(right=function_result, on='perturbation_id', how='left')
				if is_numeric(comparison['perturbed_function_result']):
					comparison['function_perturbation'] = comparison['perturbed_function_result'] - comparison['original_function_result']
				else:
					comparison['function_perturbation'] = comparison['perturbed_function_result'] != comparison['original_function_result']
				return comparison
			else:
				return perturbed__function_result



		pensieve.store(key='comparison', function=compare)
	except Exception as e:
		print(e)

	return pensieve



from .Trend import Trend
from .Nonlinear import Nonlinear

import math

class LogarithmicGrowth(Trend):
	def __init__(self, timestamps, values):
		time_shift = 1-min(timestamps)
		super().__init__(x=timestamps, y=values)
		self._time_shift = time_shift

		if self.is_decelerating(x=max(timestamps)) or self.is_decelerating(x=None):
			self._model = Nonlinear(
				transform=lambda x, c: math.log(x+time_shift) - (x+time_shift)/c
			)
			self._model.fit(x=timestamps, y=values, c0=max(timestamps) + 1)
		else:
			self._model = None

	@property
	def capacity(self):
		if self._model is None:
			return None
		else:
			return self._model.predict(x=self._model.c)[0]

	@property
	def settling_time(self):
		if self._model is None:
			return None
		else:
			return self._model.c[0]

	def predict(self, x):
		if self._model is None:
			return None
		else:
			return self._model.predict(x=min(x, self.settling_time))



from ravenclaw import is_numeric
from pandas import DataFrame
from pensieve import Pensieve


class PerturbedFeature:
	def __init__(
			self, data, output_column, input_column, output_column_type='numeric', transformation_function=None,
			perturbation_ratio=0.01
	):
		self._pensieve = Pensieve(graph_direction='LR')
		self.pensieve['data'] = data
		self.pensieve['perturbation_ratio'] = perturbation_ratio

		self.pensieve.store(
			key='transformed_data',
			function=lambda data: data.copy() if transformation_function is None else transformation_function(data),
			materialize=False,
			evaluate=False
		)

		if output_column_type != 'numeric':
			self.pensieve['output_value_counts'] = lambda transformed_data: dict(
				transformed_data[output_column].value_counts(dropna=False)
			)




	@property
	def pensieve(self):
		"""
		:rtype: Pensieve
		"""
		return self._pensieve

class Perturber:
	def __init__(
			self, data, output_column, output_column_type='numeric', function=None, input_columns=None,
			perturbation_ratio=0.01
	):
		"""
		:param DataFrame data: a data frame that should be perturbed
		:param callable function: a transformation function to be run on data after perturbation
		:param str output_column: the column whose value should be examined for the effect of perturbation
		:param int or float perturbation_ratio: the amount of perturbation as a ratio of the standard deviation of the perturbed column
		"""



		self._input_columns = input_columns or [column for column in data.columns if column != output_column]






		memory_list = []
		for column in self._input_columns:
			memory_dictionary = {'column': column}

			def get_perturbed_data(data, column, perturbation_ratio):
				if is_numeric(data[column]):
					new_data = data.copy()
					new_data[column] = data[column] + perturbation_ratio * data[column].std()
					return transformation_function(new_data)
				else:
					result = {}
					for category in set(data[column].values):
						new_data = data.copy()
						new_data[column] = category
						result[category] = transformation_function(new_data)
					return result

			perturbed_key = f'transformed_data__{column}_perturbed'

			def create_perturbed_data_function(column):
				def function(x):
					return get_perturbed_data(
						data=x['data'],
						column=column,
						perturbation_ratio=x['perturbation_ratio']
					)
				return function

			self.pensieve.store(
				key=perturbed_key,
				precursors=['data', 'perturbation_ratio'],
				function=create_perturbed_data_function(column=column),
				materialize=False,
				evaluate=False
			)

			if output_column_type == 'numeric':

				def get_mean_difference(perturbed_data, transformed_data):
					if isinstance(perturbed_data, dict):
						result = {}
						for key, data in perturbed_data.items():
							result[key] = data[output_column].mean() - transformed_data[output_column].mean()
						return result
					else:
						return perturbed_data[output_column].mean() - transformed_data[output_column].mean()

				def create_mean_difference_function(perturbed_key):
					def function(x):
						return get_mean_difference(
							perturbed_data=x[perturbed_key],
							transformed_data=x['transformed_data']
						)
					return function

				self.pensieve.store(
					key=f'mean_difference__{column}_perturbed',
					precursors=[perturbed_key, 'transformed_data'],
					function=create_mean_difference_function(perturbed_key=perturbed_key),
					evaluate=False
				)
				memory_dictionary['mean_difference'] = f'mean_difference__{column}_perturbed'

				def get_difference_ratio(mean_difference_column_perturbed, transformed_data, column):
					if isinstance(mean_difference_column_perturbed, dict):
						return None
					else:
						if mean_difference_column_perturbed == 0:
							return 0
						else:
							return mean_difference_column_perturbed / (transformed_data[column].std() * perturbation_ratio)

				def create_mean_difference_ratio(column):
					def function(x):
						return get_difference_ratio(
							mean_difference_column_perturbed=x[f'mean_difference__{column}_perturbed'],
							transformed_data=x['transformed_data'],
							column=column
						)
					return function

				self.pensieve.store(
					key=f'difference_ratio__{column}_perturbed',
					precursors=['transformed_data', f'mean_difference__{column}_perturbed'],
					function=create_mean_difference_ratio(column=column),
					evaluate=False
				)
				memory_dictionary['difference_ratio'] = f'difference_ratio__{column}_perturbed'

			else:

				def get_count_difference(perturbed_data, output_value_counts):
					if isinstance(perturbed_data, dict):
						result = {}
						for category, data in perturbed_data.items():
							perturbed_value_counts = dict(data[output_column].value_counts(dropna=False))
							count_differences = {}
							for result_category in set(output_value_counts.keys()).union(set()):
								count_differences[result_category] = perturbed_value_counts.get(result_category, 0) - \
																	 output_value_counts.get(result_category, 0)
							result[category] = count_differences
					else:
						perturbed_value_counts = dict(perturbed_data[output_column].value_counts(dropna=False))
						count_differences = {}
						for result_category in set(output_value_counts.keys()).union(set()):
							count_differences[result_category] = perturbed_value_counts.get(result_category,0) - \
																 output_value_counts.get(result_category, 0)
						return count_differences

				def create_count_difference_function(perturbed_key):
					def function(x):
						return get_count_difference(
							perturbed_data=x[perturbed_key],
							output_value_counts=x['output_value_counts'],
						)
					return function

				self.pensieve.store(
					key=f'count_difference__{column}_perturbed',
					precursors=[perturbed_key, 'transformed_data', 'output_value_counts'],
					function=create_count_difference_function(perturbed_key=perturbed_key),
					evaluate=False
				)
				memory_dictionary['count_difference'] = f'count_difference__{column}_perturbed'

			memory_list.append(memory_dictionary)

		def create_summary(memory_list):
			def function(x):
				result_list = []
				for memory_dictionary in memory_list:
					result_dictionary = {'column': memory_dictionary['column']}
					for memory_key_key, memory_key in memory_dictionary.items():
						if memory_key_key != 'column':
							result_dictionary[memory_key_key] = x[memory_key]
					result_list.append(result_dictionary)
				return result_list
			return function
		self.pensieve.store(
			key='summary',
			precursors=[
				value for memory_dictionary in memory_list for key, value in memory_dictionary.items() if key != 'column'
			],
			function=create_summary(memory_list=memory_list)
		)




import statsmodels.formula.api as smf
from statsmodels.api import families
from .Regression import Regression


class BasicGEE(Regression):
	def __init__(self, data, formula, significance_level=0.05, family=None, groups=None, model_builder=None, parent=None):
		super().__init__(
			data=data, formula=formula, model_builder=smf.gee, groups=groups, family=family,
			significance_level=significance_level, parent=parent
		)

	@property
	def summary(self):
		if self._summary is None:
			self._summary = self.fit.summary()
		return self._summary

	@property
	def coefficient_table(self):
		"""
		:rtype: DataFrame
		"""
		return super().coefficient_table.rename(columns={'P>|z|': 'p'})


class GEE(BasicGEE):
	def __init__(self, data, formula, groups=None, significance_level=0.05, model_builder=None, family=None, parent=None):
		super().__init__(
			data=data, formula=formula,  groups=groups, significance_level=significance_level, parent=parent
		)

	@property
	def name(self):
		"""
		:rtype: str
		"""
		return 'gee'

class LogisticGEE(BasicGEE):
	def __init__(self, data, formula, groups=None, significance_level=0.05, model_builder=None, family=None, parent=None):
		super().__init__(
			data=data, formula=formula, groups=groups, family=families.Binomial(),
			significance_level=significance_level, parent=parent
		)

	@property
	def name(self):
		"""
		:rtype: str
		"""
		return 'logistic_gee'



from pandas import DataFrame
import statsmodels.stats.multicomp as multicomp
from scipy import stats

class MultipleComparisons:
	def __init__(self, data, dependent_variable_column, independent_group_column):
		"""
		:param DataFrame data: input dataframe
		:param str dependent_variable_column: name of the column representing the dependent variable
		:param str independent_group_column: name of the column Comparison test should be run on
		"""
		self._independent_group_column = None
		self._dependent_variable_column = dependent_variable_column
		self._data = None
		self.independent_group_column = independent_group_column
		self.data = data


	def reset(self):
		self._comparisons = None
		self._table = None

	@property
	def dependent_variable_column(self):
		return self._dependent_variable_column

	@dependent_variable_column.setter
	def dependent_variable_column(self, dependent_variable_column):
		self._dependent_variable_column = dependent_variable_column
		self.reset()

	@property
	def independent_group_column(self):
		return self._independent_group_column

	@independent_group_column.setter
	def independent_group_column(self, group_column):
		self._independent_group_column = group_column
		self.reset()

	@property
	def data(self):
		"""
		:rtype: DataFrame
		"""
		return self._data

	@data.setter
	def data(self, data):
		self._data = data.copy()

	@property
	def comparisons(self):
		if self._comparisons is None:
			self._comparisons = multicomp.MultiComparison(
				self.data[self.dependent_variable_column],
				self.data[self.independent_group_column]
			)
		return self._comparisons

	def get_tukey_results(self, as_dataframe=True, **kwargs):
		results = self.comparisons.tukeyhsd(**kwargs)
		if not as_dataframe:
			return results
		else:
			return DataFrame(data=results._results_table.data[1:], columns=results._results_table.data[0])

	def get_bonferroni_results(self, as_dataframe=True, **kwargs):
		results = self.comparisons.allpairtest(stats.ttest_rel, method='Holm', **kwargs)
		if not as_dataframe:
			return results
		else:
			return DataFrame(data=results._results_table.data[1:], columns=results._results_table.data[0])




from .MainEffect import MainEffect
from .InteractionEffect import InteractionEffect


def Effect(formula, p=None, coefficient=None):
	"""
	:type formula: str
	:rtype: InteractionEffect or MainEffect
	"""
	if ':' in formula or '*' in formula:
		return InteractionEffect(formula=formula, p=p, coefficient=coefficient)
	else:
		return MainEffect(formula=formula, p=p, coefficient=coefficient)


###############################################################################


class Formula:
	def __init__(self, formula):
		"""
		:type formula: Formula or str
		"""
		if isinstance(formula, Formula):
			self._xs = formula._xs
			self._y = formula._y
		else:
			if '~' in formula:
				y, xs = formula.split('~')
				y, xs = y.strip(), xs.strip()
				self._y = MainEffect(formula=y)
			else:
				xs = formula
				self._y = None
			self._xs = sorted([Effect(formula=x) for x in xs.split('+')], key=lambda x: (x.num_interactions, x.name))

	@classmethod
	def from_variables(cls, independent_variables, dependent_variable=None):
		"""
		:type dependent_variable: NoneType or MainEffect
		:type independent_variables: list[MainEffect or InteractionEffect]
		:rtype: Formula
		"""
		x_formula = ' + '.join([x.formula for x in independent_variables])
		if dependent_variable is None:
			return cls(formula=x_formula)
		else:
			return cls(formula=f'{dependent_variable.formula} ~ {x_formula}')

	def represent(self, multiply_symbol='×'):
		x_formula = ' + '.join([x.represent(multiply_symbol=multiply_symbol) for x in self.effects])
		if self.dependent_variable is None:
			return x_formula
		else:
			return f'{self.dependent_variable.formula} ~ {x_formula}'

	@property
	def formula(self):
		return self.represent(multiply_symbol='*')

	def __str__(self):
		return self.represent(multiply_symbol='×')

	def __repr__(self):
		return self.__str__()

	def display(self):
		x_formula = ' + \n'.join(['\t' + x.formula for x in self.effects])
		if self.dependent_variable is None:
			return x_formula
		else:
			return f'{self.dependent_variable.formula} ~ \n{x_formula}'

	@property
	def dependent_variable(self):
		"""
		:rtype: MainEffect
		"""
		return self._y

	@property
	def effects(self):
		"""
		:rtype: list[InteractionEffect or MainEffect]
		"""
		return self._xs

	@property
	def independent_variables(self):
		"""
		:rtype: list[MainEffect]
		"""
		result = []
		for x in self._xs:
			if isinstance(x, MainEffect):
				if x not in result:
					result.append(x)
			elif isinstance(x, InteractionEffect):
				for z in x.main_effects:
					if z not in result:
						result.append(z)
		return result



# source: https://onlinecourses.science.psu.edu/stat100/node/56/
# source: http://www.dummies.com/education/math/statistics/how-to-determine-the-confidence-interval-for-a-population-proportion/

# main python libraries
import math

# idin's libraries
import slytherin.numbers as sn

# this library
from .get_z_score import get_z_score

def percent_str(x):
	return f"{round(x*100, ndigits=2)}%"

def get_population_proportion_error(group_n, sample_n, confidence=0.95):
	p = group_n/sample_n
	if p>1 or p<0: raise ValueError(f'p={p} is an invalid value.')
	standard_error = math.sqrt(p*(1-p)/sample_n)
	return standard_error*get_z_score(confidence=confidence)



class PopulationProportion:
	def __init__(self, sample_n, group_n=None, group_proportion=0.5):
		if group_n is None and group_proportion is None:
			raise ValueError('group_n and group_proportion cannot both be None.')

		if group_n is None:
			group_n = sample_n*group_proportion

		self._group_n = group_n
		self._sample_n = sample_n

	@property
	def p(self):
		return self._group_n/self._sample_n

	def get_error(self, confidence=0.95):
		return get_population_proportion_error(group_n=self._group_n, sample_n=self._sample_n, confidence=confidence)

	def get_interval_str(self, confidence=0.95):
		return f'{percent_str(self.p)} ± {percent_str(self.get_error(confidence=confidence))}'



import scipy.stats as st

# source: https://stackoverflow.com/questions/20864847/probability-to-z-score-and-vice-versa-in-python
def get_z_score(confidence):
	"""
	gets confidence interval (e.g., 0.95) and returns z-score (e.g., 1.96)
	:type confidence: float
	:rtype: float
	"""
	return st.norm.ppf(confidence/2+0.5)


from sklearn.linear_model import LinearRegression
from scipy.optimize import minimize
import numpy as np
from numpy import ndarray

class Nonlinear:
	def __init__(self, transform):
		"""
		:param callable transform: transformation function applied to x with an optional parameter c
		"""
		self._transform = transform
		self._function = None

	def fit(self, x, y, c0, optimization_method='Nelder-Mead'):
		"""
		this function fits the best function of form:
		y = a.f(x, c) + b
		:param list[float] or list[int] x: independent variable
		:param list[float] or list[int] y: dependent variable
		:param callable or NoneType transformation: a function to transform x
		:param float or int or ndarray c0: first guess for best c
		:param str optimization_method: method used for optimizing c
		:rtype: callable
		"""
		self._function = fit_nonlinear(
			x=x, y=y, transform_x=self._transform, transform_x_params=c0, optimization_method=optimization_method
		)

	train = fit

	@property
	def a(self):
		if self._function is None:
			raise RuntimeError('model is not trained yet!')
		return self._function.a

	@property
	def b(self):
		if self._function is None:
			raise RuntimeError('model is not trained yet!')
		return self._function.b

	@property
	def c(self):
		if self._function is None:
			raise RuntimeError('model is not trained yet!')
		return self._function.c

	def predict(self, x):
		if self._function is None:
			raise RuntimeError('model is not trained yet!')

		return self._function(x)


def fit_nonlinear(
		x, y,
		transform_x=None, transform_x_params=None,
		transform_y=None, transform_y_params=None,
		optimization_method='Nelder-Mead'
):
	"""
	this function fits the best function of form:
	y = a.f(x, c) + b
	:param list[float] or list[int] x: independent variable
	:param list[float] or list[int] y: dependent variable
	:param callable or NoneType transform_x: a function to transform x
	:param callabel or NoneType transform_y: a function to transform y
	:param float or int or list[float] transform_x_params: first guess for best parameters of transform_x
	:param float or int or list[float] transform_y_params: first guess for best parameters of transform_y
	:param str optimization_method: method used for optimizing
	:rtype: callable
	"""

	def get_trend_function(_x, _y, _transform_x, _transform_x_params):
		if _transform_x is None:
			u = _x.copy()
		elif _transform_x_params is None:
			u = [_transform_x(i) for i in _x]
		else:
			u = [_transform_x(i, _transform_x_params) for i in _x]

		u = np.array(u).reshape((-1, 1))
		_y = np.array(_y)

		model = LinearRegression()
		model.fit(u, _y)
		a = model.coef_[0]
		b = model.intercept_

		def func(x):
			if isinstance(x, list):
				return [a * _transform_x(x_i, _transform_x_params) + b for x_i in x]
			else:
				return a * _transform_x(x, _transform_x_params) + b

		return a, b, func

	def cost(cost_transform_x_params, cost_x, cost_y, cost_transform_x):
		a, b, f = get_trend_function(_x=cost_x, _y=cost_y, _transform_x=cost_transform_x, _transform_x_params=cost_transform_x_params)
		return sum([(cost_y_i - f(cost_x_i)) ** 2 for cost_x_i, cost_y_i in zip(cost_x, cost_y)])

	if transform_x is None:
		transform_x_params = None
	else:
		result = minimize(
			lambda c: cost(cost_transform_x_params=c, cost_x=x, cost_y=y, cost_transform_x=transform_x),
			transform_x_params,
			method=optimization_method
		)
		transform_x_params = result.x

	a, b, f = get_trend_function(_x=x, _y=y, _transform_x=transform_x, _transform_x_params=transform_x_params)
	f.a = a
	f.b = b
	f.c = transform_x_params
	return f



from setuptools import setup, find_packages

# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
	long_description = f.read()

setup(
	name='nightingale',
	version='2019.12.29',
	description='Python library for simplifying statistical analysis and making it more consistent',
	long_description=long_description,
	long_description_content_type='text/markdown',
	url='https://github.com/idin/nightingale',
	author='Idin',
	author_email='py@idin.ca',
	license='MIT',
	packages=find_packages(exclude=("jupyter_tests")),
	install_requires=[
		'statsmodels', 'scipy', 'pandas', 'slytherin', 'numpy', 'chronometry', 'pensieve', 'sklearn',
		'joblib', 'ravenclaw'
	],
	python_requires='~=3.6',
	zip_safe=False
)


from sklearn.linear_model import LinearRegression
from datetime import date, datetime
from chronometry import get_elapsed
import numpy as np

class Point:
	def __init__(self, x, y):
		if isinstance(x, (int, float)):
			self._x_type = 'numeric'
			self._reference = None
		elif isinstance(x, (datetime)):
			self._x_type = 'seconds'
			self._reference = datetime(2000, 1, 1)
		elif isinstance(x, (date)):
			self._x_type = 'days'
			self._reference = datetime(2000, 1, 1)

		else:
			raise TypeError(f'x: {x} is a {type(x).__name__} but x can be numeric, time, or date.')

		self._x = None
		self._original_x = None

		self.x = x
		self._y = y
		self._dy = None

	_STATE_VARIABLES = ['_x_type', '_reference', '_x', '_original_x', '_y', '_dy']

	def __hash__(self):
		return hash((self._original_x, self._y))

	def __getstate__(self):
		return {name: getattr(self, name) for name in self._STATE_VARIABLES}

	def __setstate__(self, state):
		for name, value in state.items():
			setattr(self, name, value)

	@property
	def x(self):
		return self._x

	@x.setter
	def x(self, x):
		self._original_x = x
		if self._x_type == 'numeric':
			self._x = x
		elif self._x_type == 'seconds':
			self._x = get_elapsed(start=self._reference, end=x, unit='second')
		elif self._x_type == 'days':
			self._x = get_elapsed(start=self._reference, end=x, unit='day')

	@property
	def y(self):
		return self._y

	@property
	def dy(self):
		return self._dy


	def __repr__(self):
		return f'({self._original_x}, {self.y})'

	def get_derivative(self, previous_point):
		"""
		:type previous_point: Point
		:rtype: float or str or None
		"""
		if self._x_type != previous_point._x_type:
			raise TypeError(f'cannot get derivative of two different x types: {self._x_type} and {previous_point._x_type}')
		if self.x == previous_point.x:
			if self.y > previous_point.y:
				return '+inf'
			elif self.y < previous_point.y:
				return '-inf'
			else:
				return None
		return (self.y - previous_point.y)/(self.x - previous_point.x)

	def __lt__(self, other):
		return self.x < other.x

	def __gt__(self, other):
		return self.x > other.x

	def __le__(self, other):
		return self.x <= other.x

	def __ge__(self, other):
		return self.x >= other.x

	def derivative_is_numeric(self):
		return self.dy is not None and not isinstance(self.dy, str)


def get_sign(x):
	if x > 0:
		return 1
	elif x < 0:
		return -1
	else:
		return 0


class Trend:
	def __init__(self, x=None, y=None, points=None):
		"""
		:type x: list[float] or NoneType
		:type y: list[float] or NoneType
		:type points: list[Point] or NoneType
		"""
		if points is None:
			if x is None or y is None:
				raise ValueError('either points or x and y should be provided!')
			points = [Point(x=x_i, y=y_i) for x_i, y_i in zip(x, y)]
			sorted_points = sorted(points)
		else:
			if x is not None or y is not None:
				raise ValueError('either points or x and y but not all of them should be provided!')
			sorted_points = sorted(points)

		x = [point.x for point in sorted_points]
		y = [point.y for point in sorted_points]

		for i in range(1, len(sorted_points)):
			point = sorted_points[i]
			previous_point = sorted_points[i - 1]
			point._dy = point.get_derivative(previous_point=previous_point)
		if len(sorted_points) > 1:
			if sorted_points[0].dy is None:
				sorted_points[0]._dy = sorted_points[1].dy

		if len(sorted_points) < 1:
			raise ValueError('trend is empty!')
		self._points = sorted_points
		self._linear = LinearRegression()
		self._derivative_linear = LinearRegression()

		if len(self._points) > 1:
			X = np.array(x).reshape((-1, 1))
			Y = np.array(y)
			self._linear.fit(X, Y)

			X2 = np.array([p.x for p in sorted_points if p.derivative_is_numeric()]).reshape((-1, 1))
			DY = np.array([p.dy for p in sorted_points if p.derivative_is_numeric()])
			self._derivative_linear.fit(X2, DY)

	def __getstate__(self):
		return {'_points': self._points, '_linear': self._linear, '_derivative_linear': self._derivative_linear}

	def __setstate__(self, state):
		for name, value in state.items():
			setattr(self, name, value)

	def __repr__(self):
		return repr(self._points)

	@property
	def points(self):
		return self._points

	@property
	def x_type(self):
		if len(self._points) > 0:
			return self._points[-1]._x_type
		else:
			return None

	def get_index(self, point):
		"""
		:type point: Point
		:rtype: int
		"""
		return self._points.index(point)

	def __getitem__(self, item):
		"""
		:type item: int
		:rtype: Point
		"""
		return self._points[item]

	@property
	def num_points(self):
		"""
		:rtype: int
		"""
		return len(self._points)

	@property
	def length(self):
		"""
		:rtype: int or float
		"""
		if self.num_points > 0:
			return self._points[-1].x - self._points[0].x
		else:
			return 0

	@property
	def sections(self):
		sections = {}
		section_number = 0
		for point in self._points:
			if point.dy is None:
				section_number += 1
			elif section_number not in sections:
				sections[section_number] = [point]
			elif get_sign(point.dy) == get_sign(sections[section_number][0].dy):
				sections[section_number].append(point)
			else:
				section_number += 1
				if section_number in sections:
					raise KeyError(f'{section_number} should not exist here!')
				sections[section_number] = [point]

		return [Trend(points=sections[section_number]) for section_number in sorted(sections.keys())]

	@property
	def minimums(self):
		minimum = min(self._points, key=lambda x: x.y)
		return [point for point in self._points if point.y == minimum.y]

	@property
	def maximums(self):
		maximum = max(self._points, key=lambda x: x.y)
		return [point for point in self._points if point.y == maximum.y]

	@property
	def a(self):
		try:
			return self._linear.coef_[0]
		except Exception as e:
			print(f'exception for:\n', self)
			raise e

	slope = a

	@property
	def b(self):
		try:
			return self._linear.intercept_
		except Exception as e:
			print(f'exception for:\n', self)
			raise e

	intercept = b

	@property
	def a_derivative(self):
		try:
			return self._derivative_linear.coef_[0]
		except Exception as e:
			print(f'exception for:\n', self)
			raise e

	@property
	def b_derivative(self):
		try:
			return self._derivative_linear.intercept_
		except Exception as e:
			print(f'exception for:\n', self)
			raise e

	def get_type(self, x=None, sensitivity=1e-6):
		if x is None:
			slope = self.a
		else:
			slope = self._derivative_linear.predict(X=x)

		if abs(slope) < sensitivity:
			if abs(self.a_derivative) < sensitivity:
				return 'stable'
			elif self.a_derivative > 0:
				return 'valley'
			else:
				return 'peak'

		elif slope > 0:
			if abs(self.a_derivative) < sensitivity:
				return 'constant_growth'
			elif self.a_derivative > 0:
				return 'accelerating_growth'
			else:
				return 'decelerating_growth'

		else:
			if abs(self.a_derivative) < sensitivity:
				return 'constant_decline'
			elif self.a_derivative > 0:
				return 'decelerating_decline'
			else:
				return 'accelerating_decline'

	def is_accelerating(self, x=None):
		return self.get_type(x=x).startswith('accelerating')

	def is_decelerating(self, x=None):
		return self.get_type(x=x).startswith('decelerating')

	def is_growing(self, x):
		return self.get_type(x=x).endswith('growth')

	def is_declining(self, x):
		return self.get_type(x=x).endswith('decline')



class RegressionError(Exception):
	pass

class FormulaError(RegressionError):
	pass

class EliminationError(RegressionError):
	pass

class AllValuesSignificantError(EliminationError):
	pass

class NoIndependentVariableError(EliminationError):
	pass


class BrokenSummary:
	def __init__(self, model):
		self._model = model

	@property
	def model(self):
		"""
		:rtype: BrokenModel
		"""
		return self._model


class BrokenModel:
	def __init__(self, exception, regression):
		"""
		:type exception: Exception
		:type regression: Regression
		"""
		self._exception = exception
		self._formula = regression.formula
		self._family = regression.family
		self._groups = regression.groups
		self._model_builder = regression._model_builder

	@property
	def exception(self):
		return self._exception

	def summary(self):
		return BrokenSummary(model=self)